from dynamod.evaluation import *
from dynamod.actionstack import *
from dynamod.checks import *
from dynamod.tracer import Tracer
from dynamod.partition import Partition
from dynamod.history import History
from collections import OrderedDict
import json
import numpy as np
import random

class DynaModel:
    def __init__(self, srcfile):
        self.srcfile = srcfile
        self.formulas = {}
        self.functions = {}
        self.parameters = {}
        self.userObjects = {}
        self.results = {}
        self.progressions = OrderedDict()
        self.autosplits = {}
        self.attDescs = OrderedDict()
        self.trace = False
        self.check = False
        self.trace_for = None
        self.raw_errors = False
        self.simulate = False
        self.ctx_stack = []
        self.error_stack = None
        self.tracer = None
        self.builtin = BuiltinFunctions(self)

    def initialize(self, parameters=None, objects=None):
        try:
            with Action(self, "initializing model", line=False):
                if parameters is not None:
                    self.parameters.update(parameters)
                if objects is not None:
                    self.userObjects.update(objects)
                self.baseStore = self.build_basestore()
                self.attSystem = AttributeSystem(self)
                self.context = DynaContext(self, Segop(self))
                for name, attDesc in self.attDescs.items():
                    self.attSystem.addAttribute(name, attDesc)
                self.context.onseg = Segop(self)
                self.all_none = [None for att in self.attSystem.attributes]
                self.matrix = np.array(self.attSystem.build_matrix())
                self.incoming = np.zeros_like(self.matrix)
                self.outgoing = np.zeros_like(self.matrix)
                self.tick = 0
                self.distributions = {}
                self.history = History(self)
                self.history.store()
                self.simulate = True
                self.step()
                self.simulate = False

        except Exception as e:
            if self.raw_errors:
                raise e from None
            report_actions(self, e)
            exit(1)

    def evalExpr (self, expr, onseg=None):
        if onseg is not None:
            self.context.onseg = onseg
        return evalExpression(expr, self.context)

    def evalCond(self, expr, onseg=None):
        if onseg is not None:
            self.context.onseg = onseg
        return evalCondition(expr, self.context)

    def tprint(self, *text):
        if self.trace:
            print (*text)

    def include(self, path):
        keep = self.srcfile
        self.srcfile = path
        parse_model(path, self)
        self.srcfile = keep

    def trace_on(self, map, split_by=None):
        self.trace_for = tuple([att.indexof(map[att.name]) if att.name in map else NOSLICE for att in self.attSystem.attributes])
        self.trace_split = None
        if split_by is not None:
            self.trace_split = 0
            for  att in self.attSystem.attributes:
                if att.name == split_by:
                    break;
                if att.name not in map:
                    self.trace_split += 1
        self.trace_val = self.get_traceval(self.matrix)

    def get_traceval(self, of):
        if self.trace_for is None:
            return None
        res = of[self.trace_for]
        if self.trace_split is None:
            return res.sum()
        return res.sum(self.trace_split)

    def attribute(self, axis):
        try:
            return self.attSystem.attr_map[axis]
        except ValueError:
            raise ConfigurationError("unknown attribute '" + axis + "'")
        
    def indexof (self, axis, value):
        att = self.attribute(axis)
        try:
            return (att.index, att.indexof(value))
        except ValueError:
            raise ConfigurationError("unknown value '" + value + "' for attribute'" + axis + "'")

    def attindex (self, axis, value):
        att = self.attribute(axis)
        try:
            return (att, att.indexof(value))
        except ValueError:
            raise ConfigurationError("unknown value '" + value + "' for attribute'" + axis + "'")

    def run(self, cycles, trace_at=None):
        for i in range(cycles):
            if i == trace_at:
                self.tracer = Tracer()
            self.step()
            if i == trace_at:
                self.tracer.finish()
                self.tracer = None

    def step(self):
        try:
            with Action(self, "progressions for tick " + str(self.tick), line=False):
                self._step()
        except MissingAxis:
            self.tprint("retry step", self.tick)
            self.trace_val = self.get_traceval(self.matrix)
            self.matrix = self.backup.copy()
            self.step()
        except Exception as e:
            if self.raw_errors:
                raise e from None
            report_actions(self, e)
            exit(1)

    def _step(self):
        self.init_step()
        for name, prog in self.progressions.items():
            with Action(self, "perform progression " + name, line=False):
                self.enter_local_context()
                if self.tracer is not None:
                    self.tracer.line("perform " + name)
                self.tprint("perform", name)
                try:
                    onsegs = self.perform_autosplit_steps (prog, name)
                    if not self.simulate:
                        self.apply_changes (onsegs)
                except MissingAxis as miss_axis:
                    raise miss_axis
                finally:
                    self.leave_local_context()
        if not self.simulate:
            self.close_step()

    def perform_autosplit_steps (self, prog, name):
        if name in self.autosplits:
            splits = self.autosplits[name]
        else:
            splits = []
        try:
            return self.perform_split_steps(prog, Segop(self), splits.copy())
        except MissingAxis as miss_axis:
            splits.append(miss_axis.axis)
            self.tprint("add axis", miss_axis, "to", name)
            self.autosplits[name] = splits
            raise miss_axis

    def perform_split_steps (self, prog, onseg:Segop, splits):
        if len(splits) == 0:
            self.baseStore.clear_cache()
            return self.perform_steps(prog, onseg)
        axis = splits.pop()
        att = self.attribute(axis)
        onsegs = []
        for seg in onseg.split_on_axis(att.index):
            onsegs.extend(self.perform_split_steps(prog, seg, splits.copy()))
        return onsegs


    def enter_local_context(self):
        self.context.values = self.context.values.extended()

    def leave_local_context(self):
        self.context.values = self.context.values.base

    def init_step(self):
        self.baseStore.clear_cache()
        self.backup = self.matrix.copy()

    def close_step(self):
        #print('.', end='')
        self.tickover()
        if self.check:
            check_correctness(self)
        if self.trace_for is not None:
            check_tickchange(self, self.tick)
        self.history.store()

    def tickover(self):
        for adist in self.distributions.values():
            adist.tickover()
        self.tick += 1

    def perform_steps(self, steps, onseg:Segop, alias=None):
        if self.tracer is not None:
            self.tracer.begin("operate on " + onseg.share_desc())
        if alias is not None:
            self.context.values.put(alias, Partition(self, onseg))
        onsegs = [onseg]
        for step in steps:
            nextsegs = []
            for seg in onsegs:
                nextsegs.extend (self.perform_step(step, seg))
            onsegs = nextsegs
        if self.tracer is not None:
            self.tracer.end()
        return onsegs

    def perform_step(self, op, onseg:Segop):
        with Action(self, op=op):
            if isinstance(op, DynamodElseList):
                return self.perform_restrictions(op, onseg)
            elif isinstance(op, DynamodAfter):
                return self.perform_after (op, onseg)
            elif isinstance(op, DynamodAction):
                return self.perform_action (op, onseg)
            elif isinstance(op, DynamodVarDef):
                self.perform_vardef (op, onseg)
                return [onseg]
            else:
                raise ConfigurationError("unknown progression operarion: " + op, op.ctx)

    def perform_action (self, op:DynamodAction, onseg: Segop):
        att = self.attribute(op.axis)
        if isinstance(op.state, str):
            if self.tracer is not None:
                self.tracer.line("set " + att.name + "=" + op.state)
            ivalue = att.indexof(op.state)
            if onseg.needs_split(att.index, ivalue):
                onsegs = onseg.split_on_att(att.index, ivalue)
                onsegs[1].set_value(att.index, ivalue)
                return onsegs
            onseg.set_value(att.index, ivalue)
            return [onseg]
        elif isinstance(op.state, dict):
            shares = {}
            for state, share in op.state.items():
                shares[att.indexof(state)] = self.evalExpr(share, onseg)
            normalize_map(shares, op.axis, op.ctx)
            if self.tracer is not None:
                self.tracer.line("set " + att.name + "=" + str(shares))
            return onseg.split_by_shares(att.index, shares)
        else:
            raise ConfigurationError("unknown state description: " + op.state, op.ctx)

    def apply_changes(self, onsegs):
        for seg in onsegs:
            if seg.change != self.all_none:
                self.apply_change(seg)
        self.matrix += (self.incoming - self.outgoing)
        self.incoming = np.zeros_like(self.matrix)
        self.outgoing = np.zeros_like(self.matrix)
        if self.check:
            check_total(self)
            check_nonnegatives(self)


    def apply_change(self, onseg):
        #key = onseg.as_key()
        for sout, sin in onseg.to_apply():
            transfer = onseg.share * self.matrix[sout]
            if transfer.sum() > 0:
                if self.trace and self.trace_for is None:
                    self.tprint(short_str(sout) + "->" + short_str(sin) + ": " + str(transfer.sum()))
                if self.tracer is not None:
                    self.tracer.line(short_str(sout) + "->" + short_str(sin) + ": " + str(transfer.sum()))
                self.outgoing[sout] += transfer
                self.incoming[sin] += transfer

                for adist in self.distributions.values():
                    adist.distribute(sin, sout, transfer, key=(deslice(sout),deslice(sin)))
        if self.check:
            check_total(self)
            check_nonnegatives(self)

    def calc_after_share(self, op, onseg):
        after = AfterDistribution.get_distribution(self, op, onseg.seg)
        return after.get_share()

    def perform_after (self, op:DynamodAfter, onseg):
        with Action(self, "perform after", op=op):
            prob = self.calc_after_share(op, onseg)
            both = onseg.split_on_prob(prob)
            onsegs = self.perform_steps(op.block, both[0])
            onsegs.append(both[1])
        return onsegs

    def perform_vardef (self, op:DynamodVarDef, onseg:Segop):
        self.context.values.put(op.varname, self.evalExpr(op.expression, onseg))

    def perform_restrictions (self, op:DynamodElseList, onseg):
        onsegs = []
        axes = set()
        axvalues = set()
        for restr in op.list:
            with Action(self, "perform conditional", op=restr):
                if restr.type == 'if':
                    is_true = self.evalCond(restr.cond, onseg)
                    if not is_true:
                        continue
                    return self.perform_steps (restr.block, onseg, restr.alias)
                elif isinstance(restr.cond, DynamodAxisValue):
                    axval = restr.cond
                    axes.add(axval.axis)
                    att = self.attribute(axval.axis)
                    if len(axes) > 1:
                        raise ConfigurationError("sequence of for-conditions must use same attribute")
                    value = axval.value
                    if isinstance(value, list):
                        axvalues.update(value)
                        for v in value:
                            ivalue = att.indexof(v)
                            if not onseg.needs_split(att.index, ivalue):
                                raise ConfigurationError("redundant or contradictive for-condition")
                            both = onseg.split_on_att(att.index, ivalue)
                            onsegs.extend (self.perform_steps(restr.block, both[0], restr.alias))
                            onseg = both[1]
                    else:
                        value = self.evalExpr(value, onseg)
                        axvalues.add(value)
                        ivalue = att.indexof(value)
                        if onseg.needs_split(att.index, ivalue):
                            both = onseg.split_on_att(att.index, ivalue)
                            onsegs.extend (self.perform_steps(restr.block, both[0], restr.alias))
                            onseg = both[1]
                        elif onseg.has_value(att.index, ivalue):
                            onsegs.extend (self.perform_steps(restr.block, onseg, restr.alias))

                else:
                    if len(axes) > 0:
                        raise ConfigurationError("cannot combine attribute and probability conditions")
                    if len(onsegs) > 0:
                        raise ConfigurationError("cannot chain probability conditions, must use 'otherwise'")
                    prob = self.evalExpr(restr.cond, onseg)
                    both = onseg.split_on_prob(prob)
                    onsegs = self.perform_steps(restr.block, both[0])
                    if op.otherwise is None:
                        onsegs.append(both[1])
                    else:
                        onsegs.extend (self.perform_steps (op.otherwise, both[1]))
                    return onsegs

        if op.otherwise is not None:
            axis = axes.pop()
            others = set(self.attribute(axis).values) - axvalues
            ivalues = [att.indexof(v) for v in others]
            splits = onseg.split_on_attlist(att.index, ivalues)
            for s in splits:
                onsegs.extend (self.perform_steps(op.otherwise, s))
        else:
            onsegs.append (onseg)
        return onsegs

    def invokeFunc(self, funcname, args):
        if self.tracer is not None:
            key = funcname + "("
            if args is not None:
                key += ", ".join(str(arg) for arg in args)
            key += ")"
            self.tracer.begin(key)
        if funcname not in self.functions:
            if hasattr(self.builtin, funcname):
                method = getattr(self.builtin, funcname)
                if callable(method):
                    res = method(*args)
            elif funcname in self.userObjects:
                method = self.userObjects[funcname]
                if callable(method):
                    res = method(*args)
            else:
                raise EvaluationError("unknown function: " + funcname)
        else:
            res = self.functions[funcname].evaluate (args, self.context)
        if self.tracer is not None:
            self.tracer.end(res)
        return res

    def build_basestore(self):
        store = GlobalStore(self)
        store = store.extendedBy(ImmutableMapStore(self.parameters))
        store = store.extendedBy(FormulaStore(self))
        store = store.extendedBy(ImmutableMapStore(self.userObjects))
        return store

    def addParameter (self, ctx, name, expr):
        if is_num(expr):
            self.parameters[name] = expr
        else:
            res = as_numlist(expr)
            if res is None:
                raise ConfigurationError("parameter '" + name + "' must be defined as a number or list of numbers")
            self.parameters[name] = res

    def addResult (self, ctx, name, expr):
        self.results[name] = DynamodExpression(self, ctx, name, expr)

    def addProgression (self, ctx, name, progressions:list, before=None):
        if before is not None:
            self.progressions = insert_at(self.progressions, name, progressions, before)
        else:
            self.progressions[name] = progressions

    def addFormula(self, ctx, name, expr):
        self.formulas[name] = DynamodExpression(self, ctx, name, expr)

    def addFunc(self, ctx, name, args, expr):
        self.functions[name] = DynamodFunction(self, ctx, name, args, expr)

    def addAttribute(self, ctx, name, attdesc:DynamodAttrib):
        self.attDescs[name] = attdesc

    def full_partition(self):
        return Partition(self)

    def get_attribute(self, axis, value, start=None, stop=None):
        return self.history.get_attribute(axis, value, start, stop)

    def get_attributes(self, axis, start=None, stop=None):
        return self.history.get_attributes(axis, start, stop)

    def get_result(self, name, start=None, stop=None):
        return self.history.get_result(name, start, stop)

    def get_results(self, names, start=None, stop=None):
        return self.history.get_results(names, start, stop)

    def get_all_results(self, start=None, stop=None):
        return self.history.get_all_results(start, stop)

class AttributeSystem:
    """holds all properties"""

    def __init__(self, model:DynaModel):
        self.model = model
        self.attributes = []
        self.attr_map = {}

    def addAttribute(self, name:str, att:DynamodAttrib):
        att = Attribute(self.model, name, att)
        att.index = len(self.attributes)
        if att.index != 0:
            self.attributes[-1].next = att
        self.attributes.append(att)
        self.attr_map[att.name] = att

    def build_matrix(self):
        given = {}
        return self.attributes[0].build_matrix (given, 1)



class Attribute:
    """one attribute used to partition a group"""

    def __init__(self, model: DynaModel, name: str, dp: DynamodAttrib):
        self.model = model
        self.index = 0
        self.next = None
        self.name = name
        self.values = dp.values
        self.value_map = {}
        for v in dp.values:
            self.value_map[v] = self.indexof(v)
        self.shares = ShareSystem (self, dp.shares)

    def indexof (self, value):
        try:
            return self.values.index(value)
        except ValueError:
            raise ConfigurationError("unknown value '" + value + "' for attribute'" + self.name + "'")

    def build_shares (self, given:dict[str,str]):
        return self.shares.build_shares (given)

    def build_matrix (self, given:dict[str,str], base):
        matrix = []
        quota = self.build_shares(given)
        i = 0
        for s in quota:
            entry = s * base
            if self.next is None:
                matrix.append(entry)
            else:
                given[self.name] = self.values[i]
                matrix.append(self.next.build_matrix (given, entry))
            i += 1
        return matrix

    def normalize_list(self, array, ctx=None):
        return normalize_list(array, self.name, ctx)

    def normalize_map(self, shares, ctx=None):
        normalize_map(shares, self.name, ctx)

class ShareSystem:
    def __init__(self, att:Attribute, shares):
        self.att = att
        self.share_list = None
        self.share_otherwise = None
        self.share_map = None
        if isinstance(shares, list):
            if len(shares) != len(att.values):
                raise ConfigurationError("share list has wrong number of entries for attribute " + att.name)
            self.share_map = {}
            for v,s in zip(att.values, shares):
                self.share_map[v] = ShareValue(att, s)
        elif isinstance(shares, dict):
            self.share_map = {}
            for v,s in shares.items():
                self.share_map[v] = ShareValue(att, s)
        elif isinstance(shares, DynamodElseList):
            self.share_list = [ConditionalShares(att, s) for s in shares.list]
            if (shares.otherwise is not None):
                self.share_otherwise = ShareSystem(att, shares.otherwise)
        else:
            raise ConfigurationError("unrecognized share system for attribute " + att.name)

    def build_shares (self, given:dict[str,str]):
        if self.share_list is None:
            array = []
            for ax in self.att.values:
                if ax not in self.share_map:
                    raise ConfigurationError("value '" + ax + "' of attribute '" + self.att.name + "' has no defined share")
                array.append(self.share_map[ax].build_share (ax, given))
            return self.att.normalize_list(array)
        for sl in self.share_list:
            if sl.matches(given):
                return sl.share.build_shares (given)
        if self.share_otherwise is None:
            raise ConfigurationError("no matching for-condition while evaluation shares for " + self.att.name + " in context " + json.dumps(given))
        return self.share_otherwise.build_shares(given)


class ConditionalShares:
    def __init__(self, att:Attribute, condshare):
        self.att = att
        if isinstance(condshare[0], DynamodAxisValue):
            self.axis = condshare[0].axis
            self.value = condshare[0].value
        else:
            raise ConfigurationError("attribute shares can only be switches by axis values")
        self.share = ShareSystem(att, condshare[1])

    def matches (self, given:dict[str,str]):
        if not self.axis in given:
            raise ConfigurationError("attribute value for '" + self.axis + "' not defined while evaluating shares of '" + self.att.name + "'")
        if isinstance(self.value, list):
            return given[self.axis] in self.value
        return given[self.axis] == self.value

class ShareValue:
    def __init__(self, att: Attribute, value):
        self.att = att
        self.share_list = None
        self.share_otherwise = None
        self.value = None
        if isinstance(value, DynamodElseList):
            self.share_list = [ConditionalShareValue(att, s) for s in value.list]
            if (value.otherwise is not None):
                self.share_otherwise = ShareValue(att, value.otherwise)
        else:
            self.value = value

    def build_share (self, axvalue:str, given:dict[str,str]):
        if self.share_list is None:
            return self.att.model.evalExpr (self.value)
        for sl in self.share_list:
            if sl.matches(given):
                return sl.share.build_share (axvalue, given)
        if self.share_otherwise is None:
            raise ConfigurationError("no matching for-condition while evaluation share'" + axvalue + "' for " + self.att.name)
        return self.share_otherwise.build_share (axvalue, given)


class ConditionalShareValue(ConditionalShares):
    def __init__(self, att:Attribute, condshare):
        self.att = att
        if isinstance(condshare.cond, DynamodAxisValue):
            self.axis = condshare.cond.axis
            self.value = condshare.cond.value
        else:
            raise ConfigurationError("attribute share can only be switches by axis values", condshare.ctx)
        self.share = ShareValue(att, condshare.expr)

class GlobalStore(ImmutableMapStore):
    def __init__(self, model):
        self.model = model
        self.here = {'ALL': self.all, 'day': self.tick, 'time': self.tick, 'random': self.random, 'PI': math.pi, 'E': math.e}

    def get(self, key):
        if key in self.here:
            obj = self.here[key]
            if callable(obj):
                return obj()
            return obj

    def all(self):
        return self.model.full_partition()

    def current(self):
        return self.model.context.segop

    def tick(self):
        return self.model.tick

    def random(self):
        return random.random()

class BuiltinFunctions:
    def __init__(self, model):
        self.model = model

    def min(self, *args):
        return min(*args)

    def max(self, *args):
        return max(*args)

    def abs(self, x):
        return abs(x)

    def ceil(self, x):
        return math.ceil(x)

    def floor(self, x):
        return math.floor(x)

    def sin(self, x):
        return math.sin(x)

    def cos(self, x):
        return math.sin(x)

    def tan(self, x):
        return math.sin(x)

    def round(self, x):
        return round(x)

class FormulaStore(ImmutableMapStore):
    def __init__(self, model):
        self.model = model
        self.here = model.formulas
        self.cache = {}

    def get(self, key):
        if key in self.cache:
            return self.cache[key]
        self.cache[key] = 0
        expr = self.here[key]
        if self.model.tracer is not None:
            self.model.tracer.begin (key)
        res = self.model.evalExpr(expr)
        if self.model.tracer is not None:
            self.model.tracer.end (res)
        self.cache[key] = res
        return res

    def clear_cache(self):
        self.cache = {}

def parse_model (srcfile:str, model=None, trace=False):
    from antlr4 import FileStream, CommonTokenStream
    from dynamod.parser.DynamodLexer import DynamodLexer
    from dynamod.parser.DynamodParser import DynamodParser
    from dynamod.builder import DynamodBuilder
    from dynamod.parse_helper import RegisterErrorListener

    input = FileStream(srcfile)
    lexer = DynamodLexer(input)
    stream = CommonTokenStream(lexer)
    parser = DynamodParser(stream)
    listener = RegisterErrorListener()
    parser.addErrorListener(listener);
    parser.setTrace(trace)
    tree = parser.model()
    if listener.had_error:
        print("processing terminated due to syntax errors in " + srcfile)
        exit(1)
    if model is None:
        model = DynaModel(srcfile)
    builder = DynamodBuilder(model)
    builder.visitModel(tree)
    return model

def get_total(matrix):
    if isinstance(matrix, list):
        return sum([get_total(s) for s in matrix])
    return matrix

def normalize_list(array, name="?", ctx=None):
    total = 0.0
    rest_at = None
    i = 0
    for s in array:
        if s == -1 and rest_at is None:
            rest_at = i
        elif s < 0:
            raise ConfigurationError("inconsistent shares for attribute " + name, ctx)
        else:
            total += s
        i += 1
    if rest_at is not None:
        array[rest_at] = 1.0 - total
        return array
    if total == 0:
        raise ConfigurationError("zero shares for attribute " + name, ctx)
    elif total < 0.999999 or total > 1.000001:
        raise ConfigurationError("shares for attribute " + name + " don't add up to 1", ctx)
    return [a/total for a in array]

def normalize_map(shares, name="?", ctx=None):
    total = 0.0
    rest_val = None
    for v, s in shares.items():
        if s == -1 and rest_val is None:
            rest_val = v
        elif s < 0:
            raise ConfigurationError("inconsistent shares for property " + name, ctx)
        else:
            total += s
    if rest_val is not None:
        shares[rest_val] = 1.0 - total
    elif total == 0:
        raise ConfigurationError("zero shares for property " + name, ctx)
    elif total < 0.9999 or total > 1.0001:
        raise ConfigurationError("shares for property " + name + " don't add up to 1", ctx)
    else:
        for v, s in shares.items():
            shares[v] = s / total




from dynamod.segop import *
import pandas as pd

class History:
    def __init__(self, model):
        self.model = model
        self.matrix = []
        self.results = {}
        for name in self.model.results:
            self.results[name] = []

    def store(self):
        self.matrix.append(self.model.matrix.copy())
        for name, expr in self.model.results.items():
            self.results[name].append (self.model.evalExpr(expr, Segop(self.model)))

    def get_attribute(self, axis, value, start=None, stop=None):
        segment = axval_segment(self.model, axis, value)
        return [m[segment].sum() for m in self.matrix[slice(start,stop,None)]]

    def get_attributes(self, axis, start=None, stop=None):
        att = self.model.attSystem.attr_map[axis]
        axl = axis_exclude(self.model, axis)
        data = [m.sum(axis=axl) for m in self.matrix[slice(start,stop,None)]]
        return pd.DataFrame(data, columns=att.values)

    def get_result(self, name, start=None, stop=None):
        pair = name.split('=')
        if len(pair) == 2:
            return self.get_attribute(pair[0], pair[1], start, stop)
        return self.results[name][slice(start,stop,None)]

    def get_results(self, names, start=None, stop=None):
        data = {name:(self.get_result(name, start, stop)) for name in names}
        return pd.DataFrame(data)

    def get_all_results(self, start=None, stop=None):
        data = {name:(hist[slice(start,stop,None)]) for (name,hist) in self.results.items()}
        return pd.DataFrame(data)



from gradescent import *
from dynamod.model import DynaModel
from dynamod.core import *
from statistics import *
import numpy as np
from numpy.polynomial import Polynomial

def minpos (xdata, ydata):
    poly = Polynomial.fit(xdata, ydata, deg=2)
    s = poly.deriv()
    x = s.roots()[0]
    return x, poly(x)
class Calibration:
    def __init__(self, model:DynaModel, cycles):
        self.model = model
        self.cycles = cycles
        self.targets = []
        self.parameters = []

    def add_target(self, resultname, expected, type='values', start=0, stop=None, weight=1, metric='mean_absolute_error'):
        self.targets.append (Target (self.model, resultname, expected, type=type, start=start, stop=stop, weight=weight, metric=metric))

    def add_variable(self, name, min=None, max=None, grid=None, is_int=False, dx=None):
        initial_value = None
        if grid is None:
            initial_value = self.model.parameters[name]
        par = Parameter(name, initial_value, min=min, max=max, grid=grid, integral=is_int, dx=dx)
        self.parameters.append(par)

    def optimize(self, zoom_limit=1e6, momentum=0, iterations=100, min_improvement=0.0001, trace=True, debug=False):
        np.seterr(all='raise')
        optimizer = Optimizer(self.get_error, args='dict', zoom_limit=zoom_limit, cfactor=0.5, momentum=momentum, iterations=iterations, min_improvement=min_improvement, trace=trace, debug=debug)
        for par in self.parameters:
            optimizer.add_par(par)
        return optimizer.optimize()

    def get_error(self, params):
        self.model.initialize(params)
        self.model.run(self.cycles)
        return sum(target.get_error() for target in self.targets)

class Target:
    def __init__(self, model:DynaModel, resultname, expected, type='values', start=0, stop=None, weight=1, metric='mean_absolute_error', series_weight=None, user_metric=None):
        self.model = model
        self.resultname = resultname
        self.expected = expected
        self.type = type
        self.start = start
        self.stop = stop
        self.weight = weight
        self.series_weight = series_weight
        self.metric = metric
        self.user_metric = user_metric

    def get_error(self):
        result = self.model.get_result(self.resultname, self.start, self.stop)
        if self.type == 'values':
            if isinstance(self.expected, list) or isinstance(self.expected, tuple):
                expected = self.expected
            else:
                try:
                    expected = [x for x in self.expected]
                except TypeError:
                    expected = [self.expected for _ in result]
            if len(result) != len(expected):
                raise ConfigurationError("number of expected values mismatch on result " + self.resultname)
        else:
            result = [self.get_minmax_point (result)]
            expected = [self.expected]
        try:
            if self.user_metric is not None:
                return self.weight * self.user_metric(result, expected)
            method = getattr(self, self.metric)
            return self.weight * method(result, expected)
        except Exception as e:
            raise ConfigurationError("error invoking metric " + self.metric) from e

    def get_minmax_point(self, values):
        find_min = self.type == 'min' or self.type == 'tmin'
        find_arg = self.type == 'tmin' or self.type == 'tmax'
        pos = np.argmin(values) if find_min else np.argmax(values)
        if pos == 0 or pos == len(values) - 1:
            return pos if find_arg else values[pos]
        poly = Polynomial.fit([pos-1, pos, pos+1], [values[pos-1], values[pos], values[pos+1]], deg=2)
        s = poly.deriv()
        x = s.roots()[0]
        return x if find_arg else poly(x)

    def mean_absolute_error(self, calced, expected):
        if self.series_weight is None:
            return mean([abs(x - y) for x, y in zip(calced, expected)])
        return mean([w * abs(x - y) for x, y, w in zip(calced, expected, self.series_weight)])


    def mean_squared_error(self, calced, expected):
        if self.series_weight is None:
            return mean([(x - y)**2 for x, y in zip(calced, expected)])
        return mean([w * (x - y)**2 for x, y, w in zip(calced, expected, self.series_weight)])

    def max_absolute_error(self, calced, expected):
        return max([abs(x - y) for x, y in zip(calced, expected)])

    def median_absolute_error(self, calced, expected):
        return median([abs(x - y) for x, y in zip(calced, expected)])




from dynamod.core import *
from dynamod.context import *
from dynamod.actionstack import Action
from dynamod.partition import Partition

class DynamodExpression:
    def __init__(self, model, ctx, name, expr):
        self.model = model
        self.srcfile = model.srcfile
        self.line = get_line(ctx)
        self.name = name
        self.expr = expr

    def evaluate(self, context:DynaContext):
        with Action(self.model, "evaluate expression " + self.name, line=self.line):
            return Evaluator(context).evalExpr(self.expr)

class DynamodFunction:
    def __init__(self, model, ctx, name, args, expr):
        self.model = model
        self.srcfile = model.srcfile
        self.line = get_line(ctx)
        self.name = name
        self.args = args
        self.expr = expr

    def evaluate(self, params, context:DynaContext):
        with Action(self.model, "evaluate function " + self.name, line=self.line):
            if not isinstance(params, list) or len(params) != len(self.args):
                raise EvaluationError("failed to invoke '" + self.name + "', wrong number of arguments", self.srcfile, self.line)
            localCtx = MapStore()
            for n,v in zip(self.args, params):
                localCtx.put(n, v)
            return Evaluator(context.chained_by(localCtx)).evalExpr(self.expr)

class Evaluator:
    def __init__(self, context:DynaContext):
        self.context = context
        self.model = context.model

    def evalComparison (self, opcode, op1, op2):
        if opcode == '<':
            return op1 < op2
        if opcode == '<=':
            return op1 <= op2;
        if opcode == '>':
            return op1 > op2
        if opcode == '>=':
            return op1 >= op2;
        if opcode == '==':
            return op1 == op2
        if opcode == '!=':
            return op1 != op2;
        raise ConfigurationError("unknown comparison operator: " + opcode)

    def evalCalculation(self, opcode, op1, op2):
        if opcode == '+':
            return op1 + op2
        if opcode == '-':
            return op1 - op2;
        if opcode == '*':
            return op1 * op2
        if opcode == '/':
            if op2 == 0:
                return op1
            return op1 / op2;
        if opcode == '**':
            return op1 ** op2;
        raise ConfigurationError("unknown calculation operator: " + opcode)

    def evalCond (self, expr):
        with Action(self.model, "evaluate condition", op=expr):
            if isinstance(expr, UnaryOp):
                if expr.opcode == 'or':
                    for sub in expr.op:
                        if self.evalCond (sub):
                            return True
                    return False
                if expr.opcode == 'and':
                    for sub in expr.op:
                        if not self.evalCond (sub):
                            return False
                    return True
                if expr.opcode == 'not':
                    return not self.evalCond (expr.op)
                raise ConfigurationError("unknown condition operation(1): " + expr.opcode, expr.ctx)

            if isinstance(expr, BinaryOp):
                op1 = self.evalExpr(expr.op1)
                op2 = self.evalExpr(expr.op2)
                if is_num(op1) and is_num(op2):
                    return self.evalComparison (expr.opcode, op1, op2)
                raise ConfigurationError("illegal comparision", expr.ctx)

            if isinstance(expr, TernaryOp):
                if expr.opcode == 'between':
                    val = self.evalExpr(expr.op1)
                    limFrom = self.evalExpr(expr.op2)
                    limTo = self.evalExpr(expr.op3)
                    if is_num(val) and is_num(limFrom) and is_num(limTo):
                        return self.evalComparison('>=', val, limFrom) and self.evalComparison('<=', val, limTo)
                    raise ConfigurationError("illegal comparision", expr.ctx)
                raise ConfigurationError("unknown condition operation(3): " + expr.opcode, expr.ctx)

            raise ConfigurationError("unknown condition rule:" + str(expr))

    def evalExpr (self, expr):
        if isinstance(expr, int) or isinstance(expr, float) or isinstance(expr, str):
            return expr

        with Action(self.model, "evaluate expression", op=expr):
            if isinstance(expr, UnaryOp):
                if expr.opcode == 'var':
                    if self.context.values.knows (expr.op):
                        return self.context.values.get (expr.op)
                    raise ConfigurationError("unknown variable: " + expr.op, expr.ctx)
                if expr.opcode == 'list':
                    return [self.evalExpr(x) for x in expr.op]
                raise ConfigurationError("unknown expression type(1): " + expr.opcode, expr.ctx)

            if isinstance(expr, BinaryOp):
                if expr.opcode in ['+', '-', '*', '/', '**']:
                    if self.model.tracer is not None:
                        self.model.tracer.begin({'+':'add', '-':'subtract', '*':'multiply', '/':'divide', '**':'exp'}[expr.opcode])
                    op1 = self.evalExpr(expr.op1)
                    op2 = self.evalExpr(expr.op2)
                    if is_num(op1) and is_num(op2):
                        res = self.evalCalculation (expr.opcode, op1, op2)
                        if self.model.tracer is not None:
                            self.model.tracer.end(str(op1) + " " + expr.opcode + " " + str(op2) + " = " + str(res))
                        return res
                    raise ConfigurationError("illegal calculation operands: " + str(op1) + expr.opcode + str(op2), expr.ctx)
                if expr.opcode == 'func':
                    args = []
                    if expr.op2 is not None:
                        for op in expr.op2:
                            args.append(self.evalExpr(op))
                    return self.model.invokeFunc (expr.op1, args)
                if expr.opcode == 'dot':
                    op1 = self.evalExpr(expr.op1)
                    if isinstance(op1, Partition):
                        att = self.model.attribute(expr.op2)
                        return op1.onseg.get_value(att.index)
                    if hasattr(op1, expr.op2):
                        return getattr(op1, expr.op2)
                    raise ConfigurationError("unknown field " + expr.op2, expr.ctx)
                if expr.opcode == 'index':
                    op1 = self.evalExpr(expr.op1)
                    op2 = self.evalExpr(expr.op2)
                    return op1[op2]
                if expr.opcode == 'with':
                    if expr.op1 is None:
                        if self.context.values.knows('_SHARE_BASE'):
                            part = self.context.values.get('_SHARE_BASE')
                        else:
                            part = self.model.full_partition()
                    else:
                        part = self.evalExpr(expr.op1)
                        if not isinstance(part, Partition):
                            raise ConfigurationError("base of with-operator must be a partition", expr.ctx)
                    axval = expr.op2
                    if isinstance(axval.value, list):
                        value = [self.evalExpr(v) for v in axval.value]
                    else:
                        value = self.evalExpr(axval.value)
                    return part.restricted(axval.axis, value)
                if expr.opcode == 'share':
                    if expr.op2 is not None:
                        base = self.evalExpr(expr.op2)
                        if not isinstance(base, Partition):
                            raise ConfigurationError("base of [...|...] must be a partition", expr.ctx)
                        self.context.values.put('_SHARE_BASE', base)
                    part = self.evalExpr(expr.op1)
                    if not isinstance(part, Partition):
                        raise ConfigurationError("expression in [...] must be a partition", expr.ctx)
                    if expr.op2 is not None:
                        self.context.values.delete('_SHARE_BASE')
                    share = part.total()
                    if share > 0 and expr.op2 is not None:
                        share /= base.total()
                    return share

                raise ConfigurationError("unknown expression operation(2): " + expr.opcode, expr.ctx)

            if isinstance(expr, TernaryOp):
                if expr.opcode == 'method':
                    args = []
                    if expr.op3 is not None:
                        for op in expr.op3:
                            args.append(self.evalExpr(op))
                    obj = self.evalExpr(expr.op1)
                    methodname = expr.op2
                    if hasattr(obj, methodname):
                        method = getattr(obj, methodname)
                        if callable(method):
                            return method(*args)
                    raise ConfigurationError("unknown method '" + methodname + "' on " + str(obj), expr.ctx)

                raise ConfigurationError("unknown expression operation(3): " + expr.opcode, expr.ctx)

            if isinstance(expr, DynamodElseList):
                return self.eval_restrictions(expr)

            if isinstance(expr, DynamodExpression):
                return self.evalExpr(expr.expr)

            raise ConfigurationError("unknown expression rule: " + str(expr))

    def eval_restrictions (self, op:DynamodElseList):
        with Action(self.model, "evaluate conditional expression", op=op):
            for cond_expr in op.list:
                if cond_expr.type == 'if':
                    if self.evalCond(cond_expr.cond):
                        return self.evalExpr (cond_expr.expr)
                elif isinstance(cond_expr, DynamodCondExp):
                    axis = cond_expr.cond.axis
                    att = self.model.attribute(axis)
                    value = cond_expr.cond.value
                    if isinstance(value, list):
                        values = [att.indexof(self.evalExpr(v)) for v in value]
                        if self.context.onseg.has_segments(att.index, values):
                            return self.evalExpr (cond_expr.expr)
                    else:
                        value = self.evalExpr(value)
                        if self.context.onseg.has_segment(att.index, att.indexof(value)):
                            return self.evalExpr(cond_expr.expr)
                else:
                    raise EvaluationError("illegal conditional expression" + str(cond_expr.cond))

            if op.otherwise is None:
                raise EvaluationError("missing 'otherwise' clause")
            return self.evalExpr(op.otherwise)

def evalExpression (expr, context):
    return Evaluator(context).evalExpr(expr)

def evalCondition (expr, context):
    return Evaluator(context).evalCond(expr)




from recordtype import recordtype
from collections import OrderedDict
import random
import string

NOSLICE = slice(None)

def is_num (expr):
    return isinstance(expr, int) or isinstance(expr, float)

def as_numlist (expr):
    res = []
    if isinstance(expr, UnaryOp) and expr.opcode == 'list':
        for arg in expr.op:
            if not is_num(arg):
                return None
            res.append(arg)
        return res
    return None

def get_line(ctx):
    if ctx is not None:
        sym = None
        if hasattr(ctx, 'symbol'):
            sym = getattr(ctx, 'symbol')
        elif hasattr(ctx, 'start'):
            sym = getattr(ctx, 'start')
        if hasattr(sym, 'line'):
           return sym.line

def get_random_string(length):
    return ''.join(random.choice(string.ascii_letters) for i in range(length))

def deslice(org):
    return tuple([None if x is NOSLICE else x for x in org])

def reslice(org):
    return tuple([NOSLICE if x is None else x for x in org])

def short_str (seg):
    return '[' + ','.join(':' if s is None or s is NOSLICE else str(s) for s in seg) + ']'

def leads_into(seg, sout, sin):
    outside = False
    for (ihere, ifrom, ito) in zip(seg, sout, sin):
        #rules: no lists in ifrom and ito
        #       if one of ifrom/ito is NOSLICE, then the other one is NOSLICE too
        if ihere is not NOSLICE and ifrom is not NOSLICE:
            if ifrom == ito:
                if ihere != ifrom:
                    return False    #doesn't lead in
            else:
                if isinstance(ihere, list):
                    if ito not in ihere:
                        return False    #doesn't lead in
                    if ifrom not in ihere:
                        outside = True       #may lead out
                else:
                    if ito != ihere:
                        return False    #doesn't lead in
                    if ifrom != ihere:
                        outside = True      #may lead out
    return outside

def intersect(seg1, seg2):
    res = []
    sub1 = []
    sub2 = []
    for s1,s2 in zip(seg1, seg2):
        if s1 is NOSLICE:
            if s2 is NOSLICE:
                res.append(NOSLICE)
                sub1.append(NOSLICE)
                sub2.append(NOSLICE)
            else:
                res.append(s2)
                sub1.append(s2)
        else:
            if s2 is NOSLICE:
                res.append(s1)
                sub2.append(s1)
            elif s1 == s2:
                res.append(s1)
            else:
                return (None, None, None)
    return (tuple(res), tuple(sub1), tuple(sub2))

def axis_exclude (model, axis):
    return tuple([att.index for att in model.attSystem.attributes if att.name != axis])

def axval_segment (model, axis, value):
    return tuple([att.values.index(value) if att.name == axis else NOSLICE for att in model.attSystem.attributes])

def modified_seg(seg, index, value):
    return tuple([value if i==index else seg[i] for i in range(len(seg))])

def insert_at (map:dict, key, value, at):
    res = OrderedDict()
    found = False
    for k, v in map.items():
        if k == at:
            res[key] = value
            found = True
        res[k] = v
    if not found:
        res[key] = value
    return res

class MissingAxis(Exception):
    def __init__(self, axis):
        self.axis = axis

class ConfigurationError(Exception):
    def __init__(self, message, ctx=None, line=None, column=None):
        if ctx is not None:
            sym = None
            if hasattr(ctx, 'symbol'):
                sym = getattr(ctx, 'symbol')
            elif hasattr(ctx, 'start'):
                sym = getattr(ctx, 'start')
            if hasattr(sym, 'line'):
                line = sym.line
            if hasattr(sym, 'column'):
                column = sym.column
        if line is not None:
            if column is not None:
                message = '[' + str(line) + ',' + str(column) + '] ' + message
            else:
                message = '[' + str(line) + '] ' + message
        self.message = message

class EvaluationError(ConfigurationError):
    def __init__(self, message, srcfile=None, line=None):
        if srcfile is not None and line is not None:
            message = '[' + srcfile + ':' + line + '] ' + message
        self.message = message


DynamodDesc = recordtype('DynamodDesc', ['basis', 'params', 'properties', 'formulas', 'progressions', 'results'], default=None)
DynamodAttrib = recordtype('DynamodAttrib', ['values', 'shares'])
DynamodAxisValue  = recordtype('DynamodAxisValue', ['ctx', 'axis', 'value'])
DynamodFormula  = recordtype('DynamodFormula', ['ctx', 'name', 'args', 'expr'])
DynamodElseList = recordtype('DynamodElseList', ['ctx', 'list', 'otherwise'])
DynamodVarDef = recordtype('DynamodVarDef', ['ctx', 'varname', 'expression'])
DynamodAfter = recordtype('DynamodAfter', ['ctx', 'distrib', 'args', 'block', 'key'])
DynamodAction = recordtype('DynamodAction', ['ctx', 'axis', 'state'])
DynamodRestriction = recordtype('DynamodRestriction', ['ctx', 'type', 'cond', 'block', ('alias',None)])
DynamodCondExp = recordtype('DynamodCondExp', ['ctx', 'type', 'cond', 'expr'])

TernaryOp = recordtype('TernaryOp', ['ctx', 'opcode', 'op1', 'op2', 'op3'])
BinaryOp = recordtype('BinaryOp', ['ctx', 'opcode', 'op1', 'op2'])
UnaryOp = recordtype('UnaryOp', ['ctx', 'opcode', 'op'])




# Generated from Dynamod.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .DynamodParser import DynamodParser
else:
    from DynamodParser import DynamodParser

# This class defines a complete generic visitor for a parse tree produced by DynamodParser.

class DynamodVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by DynamodParser#model.
    def visitModel(self, ctx:DynamodParser.ModelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_inc.
    def visitModel_inc(self, ctx:DynamodParser.Model_incContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_pars.
    def visitModel_pars(self, ctx:DynamodParser.Model_parsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_attribs.
    def visitModel_attribs(self, ctx:DynamodParser.Model_attribsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_formulas.
    def visitModel_formulas(self, ctx:DynamodParser.Model_formulasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_progressions.
    def visitModel_progressions(self, ctx:DynamodParser.Model_progressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_results.
    def visitModel_results(self, ctx:DynamodParser.Model_resultsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#model_ignore.
    def visitModel_ignore(self, ctx:DynamodParser.Model_ignoreContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#parameters.
    def visitParameters(self, ctx:DynamodParser.ParametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#parameter.
    def visitParameter(self, ctx:DynamodParser.ParameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#attributes.
    def visitAttributes(self, ctx:DynamodParser.AttributesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#attribute.
    def visitAttribute(self, ctx:DynamodParser.AttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#attribute_block.
    def visitAttribute_block(self, ctx:DynamodParser.Attribute_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#shares_as_list.
    def visitShares_as_list(self, ctx:DynamodParser.Shares_as_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#shares_as_map.
    def visitShares_as_map(self, ctx:DynamodParser.Shares_as_mapContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#shares_as_cond.
    def visitShares_as_cond(self, ctx:DynamodParser.Shares_as_condContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#share_map_block.
    def visitShare_map_block(self, ctx:DynamodParser.Share_map_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#share_map.
    def visitShare_map(self, ctx:DynamodParser.Share_mapContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#share_def.
    def visitShare_def(self, ctx:DynamodParser.Share_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#cond_shares_block.
    def visitCond_shares_block(self, ctx:DynamodParser.Cond_shares_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#cond_shares.
    def visitCond_shares(self, ctx:DynamodParser.Cond_sharesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#cond_share.
    def visitCond_share(self, ctx:DynamodParser.Cond_shareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexpr_as_simple.
    def visitPexpr_as_simple(self, ctx:DynamodParser.Pexpr_as_simpleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexpr_as_block.
    def visitPexpr_as_block(self, ctx:DynamodParser.Pexpr_as_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexpression_block.
    def visitPexpression_block(self, ctx:DynamodParser.Pexpression_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexp_list.
    def visitPexp_list(self, ctx:DynamodParser.Pexp_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexp_for.
    def visitPexp_for(self, ctx:DynamodParser.Pexp_forContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pexp_if.
    def visitPexp_if(self, ctx:DynamodParser.Pexp_ifContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#seg_as_axval.
    def visitSeg_as_axval(self, ctx:DynamodParser.Seg_as_axvalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#seg_as_eq.
    def visitSeg_as_eq(self, ctx:DynamodParser.Seg_as_eqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#seg_as_in.
    def visitSeg_as_in(self, ctx:DynamodParser.Seg_as_inContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#values.
    def visitValues(self, ctx:DynamodParser.ValuesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expression_list.
    def visitExpression_list(self, ctx:DynamodParser.Expression_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progressions.
    def visitProgressions(self, ctx:DynamodParser.ProgressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progression.
    def visitProgression(self, ctx:DynamodParser.ProgressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progression_block.
    def visitProgression_block(self, ctx:DynamodParser.Progression_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progression_statements.
    def visitProgression_statements(self, ctx:DynamodParser.Progression_statementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#prog_vardef.
    def visitProg_vardef(self, ctx:DynamodParser.Prog_vardefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#prog_restrictions.
    def visitProg_restrictions(self, ctx:DynamodParser.Prog_restrictionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#prog_after.
    def visitProg_after(self, ctx:DynamodParser.Prog_afterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#prog_action.
    def visitProg_action(self, ctx:DynamodParser.Prog_actionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#restr_for.
    def visitRestr_for(self, ctx:DynamodParser.Restr_forContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#restr_prob.
    def visitRestr_prob(self, ctx:DynamodParser.Restr_probContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#restr_if.
    def visitRestr_if(self, ctx:DynamodParser.Restr_ifContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progression_after.
    def visitProgression_after(self, ctx:DynamodParser.Progression_afterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#progression_action.
    def visitProgression_action(self, ctx:DynamodParser.Progression_actionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pstate_name.
    def visitPstate_name(self, ctx:DynamodParser.Pstate_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#pstate_block.
    def visitPstate_block(self, ctx:DynamodParser.Pstate_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#results.
    def visitResults(self, ctx:DynamodParser.ResultsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#result.
    def visitResult(self, ctx:DynamodParser.ResultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#formulas.
    def visitFormulas(self, ctx:DynamodParser.FormulasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#formula_expr.
    def visitFormula_expr(self, ctx:DynamodParser.Formula_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#formula_func.
    def visitFormula_func(self, ctx:DynamodParser.Formula_funcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#formal_args.
    def visitFormal_args(self, ctx:DynamodParser.Formal_argsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#variable_definition.
    def visitVariable_definition(self, ctx:DynamodParser.Variable_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expr_ifelse.
    def visitExpr_ifelse(self, ctx:DynamodParser.Expr_ifelseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expr_value.
    def visitExpr_value(self, ctx:DynamodParser.Expr_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#disj_ors.
    def visitDisj_ors(self, ctx:DynamodParser.Disj_orsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#disj_one.
    def visitDisj_one(self, ctx:DynamodParser.Disj_oneContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#conj_ands.
    def visitConj_ands(self, ctx:DynamodParser.Conj_andsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#conj_comp.
    def visitConj_comp(self, ctx:DynamodParser.Conj_compContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#comp_two_ops.
    def visitComp_two_ops(self, ctx:DynamodParser.Comp_two_opsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#comp_not.
    def visitComp_not(self, ctx:DynamodParser.Comp_notContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#comp_interval.
    def visitComp_interval(self, ctx:DynamodParser.Comp_intervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expval_sub.
    def visitExpval_sub(self, ctx:DynamodParser.Expval_subContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expval_add.
    def visitExpval_add(self, ctx:DynamodParser.Expval_addContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#expval_term.
    def visitExpval_term(self, ctx:DynamodParser.Expval_termContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#term_mul.
    def visitTerm_mul(self, ctx:DynamodParser.Term_mulContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#term_exp.
    def visitTerm_exp(self, ctx:DynamodParser.Term_expContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#term_factor.
    def visitTerm_factor(self, ctx:DynamodParser.Term_factorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#term_div.
    def visitTerm_div(self, ctx:DynamodParser.Term_divContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_pos.
    def visitFactor_pos(self, ctx:DynamodParser.Factor_posContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_neg.
    def visitFactor_neg(self, ctx:DynamodParser.Factor_negContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_primary.
    def visitFactor_primary(self, ctx:DynamodParser.Factor_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_expr.
    def visitFactor_expr(self, ctx:DynamodParser.Factor_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_number.
    def visitFactor_number(self, ctx:DynamodParser.Factor_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_percent.
    def visitFactor_percent(self, ctx:DynamodParser.Factor_percentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#factor_rest.
    def visitFactor_rest(self, ctx:DynamodParser.Factor_restContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_string.
    def visitPrimary_string(self, ctx:DynamodParser.Primary_stringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_func.
    def visitPrimary_func(self, ctx:DynamodParser.Primary_funcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_method.
    def visitPrimary_method(self, ctx:DynamodParser.Primary_methodContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_dot.
    def visitPrimary_dot(self, ctx:DynamodParser.Primary_dotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_list.
    def visitPrimary_list(self, ctx:DynamodParser.Primary_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_partition.
    def visitPrimary_partition(self, ctx:DynamodParser.Primary_partitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_share.
    def visitPrimary_share(self, ctx:DynamodParser.Primary_shareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_name.
    def visitPrimary_name(self, ctx:DynamodParser.Primary_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_rel_share.
    def visitPrimary_rel_share(self, ctx:DynamodParser.Primary_rel_shareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#primary_indexed.
    def visitPrimary_indexed(self, ctx:DynamodParser.Primary_indexedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#part_expr.
    def visitPart_expr(self, ctx:DynamodParser.Part_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#part_method.
    def visitPart_method(self, ctx:DynamodParser.Part_methodContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#part_segment.
    def visitPart_segment(self, ctx:DynamodParser.Part_segmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#part_name.
    def visitPart_name(self, ctx:DynamodParser.Part_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#part_with.
    def visitPart_with(self, ctx:DynamodParser.Part_withContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynamodParser#arguments.
    def visitArguments(self, ctx:DynamodParser.ArgumentsContext):
        return self.visitChildren(ctx)



del DynamodParser


from dynamod.afterdist import *

def check_nonnegatives(model):
    matrix = model.matrix + model.incoming - model.outgoing
    if np.amin(matrix) < -0.00000001:
        at = np.unravel_index(np.argmin(matrix), matrix.shape)
        print ("!!! negative share for ", end='')
        for att, index in zip(model.attSystem.attributes, at):
            print (att.name + ":" + att.values[index], end=', ')
        print(" entrycode", at)
        raise EvaluationError("negative matrix entries")

def check_total(model):
    total = (model.matrix + model.incoming - model.outgoing).sum()
    if total < 0.99999999 or total > 1.00000001:
        print ("!!! total is ", total)
        raise EvaluationError("inconsistent matrix total")

def check_changes(model, sin, sout, transfer):
    mysin, tr_subin, my_subin = intersect(sin, model.trace_for)
    mysout, tr_subout, my_subout = intersect(sout, model.trace_for)
    if mysin is not None and mysout is None:
        print ("increase by:", transfer[tr_subin].sum())
    if mysout is not None and mysin is None:
        print ("decrease by:", transfer[tr_subout].sum())

def check_tickchange(model, tick):
    matrix = model.matrix
    value = model.get_traceval(matrix)
    if value != model.trace_val:
        print("[" + str(tick) + "] value changed from", model.trace_val, "to", value)
        model.trace_val = value

def check_correctness(model):
    #matrix = model.matrix + model.incoming + model.outgoing
    if abs(model.matrix.sum() - 1) > 0.00000001:
        raise EvaluationError("matrix total is " + str(model.matrix.sum()))



from dynamod.core import MissingAxis, ConfigurationError, get_line

class Action:

    def __init__(self, model, desc=None, op=None, ctx=None, line=None):
        self.model = model
        self.desc = desc
        if op is not None and hasattr(op, "ctx"):
            ctx = op.ctx
        if ctx is not None:
            line = get_line (ctx)
        self.line = line

    def __enter__(self):
        self.model.ctx_stack.append (self)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.model.error_stack is None and exc_type is not None and exc_type != MissingAxis:
            self.model.error_stack = self.model.ctx_stack.copy()
        self.model.ctx_stack.pop()

    def to_text(self):
        return (self.desc if self.desc is not None else "error") + ((" in line " + str(self.line)) if self.line is not None and self.line is not False else "")

def report_actions (model, e):
    lines = set()
    print ("an error occurred:" + (e.message if hasattr(e, 'message') else str(e.args)))
    print ("logical operation stack:")
    if model.error_stack is not None:
        for action in model.error_stack[::-1]:
            if action.line == None:
                continue
            if action.line == False or action.line not in lines:
                print ("  ***", action.to_text())
                lines.add(action.line)
        model.error_stack = None
    if not isinstance(e, ConfigurationError):
        raise e from None
